{
TGMMarker component

  ES: componente para mostrar marcadores en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to show markers on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner los
      marcadores a mostrar
  EN: put the component into a form, link to a TGMMap and put the markers to show
=========================================================================
History:

ver 0.1.4
  ES:
    cambio: cambio en el JavaScript de algunos mtodos Set
  EN:
    change: JavaScript changed from some Set methods

ver 0.1.3
  ES:
    nuevo: aadido propiedad Icon para especificar la imagen del marcador. Puede
      ser una imagen en la web (url) o una imagen en el ordenador. En blanco ser
      la imagen por defecto.
    nuevo: aadido evento OnIconChange
    cambio: mtodos Set cambiados para evitar duplicidad de cdigo
  EN:
    new: added Icon property to specify a marker image. You can specify a web
      image (url) or a file in the PC. Nothing for default image
    new: added OnIconChange event
    change: changed Set methods to avoid duplicate code

ver 0.1.2
  ES:
    error: algunas propiedades daban un error de JavaScript al modificarse
  EN:
    bug: some properties giving a JavaScript error when modifying

ver 0.1.1
  ES:
    nuevo: Aadido evento Assign a la clase TMarkerLatLng
    cambio: Cuando se cambia la lat/lng de un marcador, ste cambia automticamente en el mapa
  EN:
    new: Added Assign Assign to the TMarkerLatLng class
    change: When a lat/lng is changed into a marker, this is automatically changed into the map

ver 0.1:
  ES- primera versin
  EN- first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMMarker;

interface

uses
  Classes,
  GMMap, GMLatLng;

type
  TGMMarker = class;
  TMarker = class;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TAnimation = class(TPersistent)
  private
    FMarker: TMarker;
    FOnDrop: Boolean;
    FBounce: Boolean;
    FIdxList: Cardinal;
    procedure SetOnDrop(const Value: Boolean);
  protected
    procedure SetBounce(const Value: Boolean); virtual;

    property IdxList: Cardinal read FIdxList write FIdxList;
  public
    constructor Create(aOwner: TMarker); virtual;

    procedure Assign(Source: TPersistent); override;
  published
    property OnDrop: Boolean read FOnDrop write SetOnDrop;
    property Bounce: Boolean read FBounce write SetBounce;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TMarkerLatLng = class(TLatLng)
  private
    FMarker: TMarker;
    FIdxList: Cardinal;
    procedure PositionChange;
  protected
    procedure SetLat(const Value: Real); override;
    procedure SetLng(const Value: Real); override;

    property IdxList: Cardinal read FIdxList write FIdxList;
  public
    constructor Create(Marker: TMarker; Lat: Real = 0; Lng: Real = 0); reintroduce; virtual;

    procedure Assign(Source: TPersistent); override;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TMarker = class(TLinkedComponent)
  private
    FAnimation: TAnimation;
    FClickable: Boolean;
    FDraggable: Boolean;
    FFlat: Boolean;
    FOptimized: Boolean;
    FPosition: TMarkerLatLng;
    FRaiseOnDrag: Boolean;
    FTitle: string;
    FVisible: Boolean;
    FIcon: string;
    procedure SetClickable(const Value: Boolean);
    procedure SetDraggable(const Value: Boolean);
    procedure SetFlat(const Value: Boolean);
    procedure SetRaiseOnDrag(const Value: Boolean);
    procedure SetVisible(const Value: Boolean);
    function GetZIndex: Integer;
    procedure SetOptimized(const Value: Boolean);
    procedure SetTitle(const Value: string);
    procedure SetIcon(const Value: string);
  protected
    procedure SetIdxList(const Value: Cardinal); override;

    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure CenterMapToMarker;
  published
    property Animation: TAnimation read FAnimation write FAnimation;
    property Clickable: Boolean read FClickable write SetClickable;  // slo regenera el propio Marker
    property Draggable: Boolean read FDraggable write SetDraggable;  // slo regenera el propio Marker
    property Flat: Boolean read FFlat write SetFlat;                 // slo regenera el propio Marker
    property Position: TMarkerLatLng read FPosition write FPosition; // slo regenera el propio Marker
    property Title: string read FTitle write SetTitle;               // slo regenera el propio Marker
    property Visible: Boolean read FVisible write SetVisible;        // slo regenera el propio Marker
    property Optimized: Boolean read FOptimized write SetOptimized;       // regenera todos los Markers
    property RaiseOnDrag: Boolean read FRaiseOnDrag write SetRaiseOnDrag; // regenera todos los Markers
    property Icon: string read FIcon write SetIcon;
    property ZIndex: Integer read GetZIndex;
    property InfoWindow;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TMarkers = class(TLinkedComponents)
  private
    procedure SetItems(I: Integer; const Value: TMarker);
    function GetItems(I: Integer): TMarker;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TMarker;
    function Insert(Index: Integer): TMarker;

    property Items[I: Integer]: TMarker read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Marker
  **************************************************************************** }
  TGMMarker  = class(TGMLinkedComponent)
  private
    FOnClick: TLatLngIdxEvent;
    FOnDblClick: TLatLngIdxEvent;
    FOnDrag: TLatLngIdxEvent;
    FOnDragEnd: TLatLngIdxEvent;
    FOnRightClick: TLatLngIdxEvent;
    FOnMouseDown: TLatLngIdxEvent;
    FOnMouseUp: TLatLngIdxEvent;
    FOnMouseOut: TLatLngIdxEvent;
    FOnDragStart: TLatLngIdxEvent;
    FOnMouseOver: TLatLngIdxEvent;
    FOnClickableChange: TLinkedComponentChange;
    FOnDragableChange: TLinkedComponentChange;
    FOnFlatChange: TLinkedComponentChange;
    FOnPositionChange: TLatLngIdxEvent;
    FOnTitleChange: TLinkedComponentChange;
    FOnVisibleChange: TLinkedComponentChange;
    FOnIconChange: TLinkedComponentChange;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TMarker;

    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add(Lat: Real = 0; Lng: Real = 0; Title: string = ''): TMarker;

    property Items[I: Integer]: TMarker read GetItems; default;
  published
    property VisualObjects;
    // eventos
    // events
        // from map
    property OnClick: TLatLngIdxEvent read FOnClick write FOnClick;
    property OnDblClick: TLatLngIdxEvent read FOnDblClick write FOnDblClick;
    property OnDrag: TLatLngIdxEvent read FOnDrag write FOnDrag;
    property OnDragEnd: TLatLngIdxEvent read FOnDragEnd write FOnDragEnd;
    property OnDragStart: TLatLngIdxEvent read FOnDragStart write FOnDragStart;
    property OnMouseDown: TLatLngIdxEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseOut: TLatLngIdxEvent read FOnMouseOut write FOnMouseOut;
    property OnMouseOver: TLatLngIdxEvent read FOnMouseOver write FOnMouseOver;
    property OnMouseUp: TLatLngIdxEvent read FOnMouseUp write FOnMouseUp;
    property OnRightClick: TLatLngIdxEvent read FOnRightClick write FOnRightClick;
        // from properties
    property OnClickableChange: TLinkedComponentChange read FOnClickableChange write FOnClickableChange;
    property OnDragableChange: TLinkedComponentChange read FOnDragableChange write FOnDragableChange;
    property OnFlatChange: TLinkedComponentChange read FOnFlatChange write FOnFlatChange;
    property OnPositionChange: TLatLngIdxEvent read FOnPositionChange write FOnPositionChange;
    property OnTitleChange: TLinkedComponentChange read FOnTitleChange write FOnTitleChange;
    property OnVisibleChange: TLinkedComponentChange read FOnVisibleChange write FOnVisibleChange;
    property OnIconChange: TLinkedComponentChange read FOnIconChange write FOnIconChange;
  end;

implementation

uses
  SysUtils,
  WebControl, GMConstants, Lang;

{ TGMMarker }

function TGMMarker.Add(Lat, Lng: Real; Title: string): TMarker;
begin
  Result := TMarker(inherited Add);
  Result.Position.Lat := Lat;
  Result.Position.Lng := Lng;
  Result.Title := Title;
end;

function TGMMarker.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Marker';
end;

function TGMMarker.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TMarkers;
end;

function TGMMarker.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TMarker;
end;

function TGMMarker.GetItems(I: Integer): TMarker;
begin
  Result := TMarker(inherited Items[i]);
end;

procedure TGMMarker.ShowElements;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
  i: Integer;
  Icon: string;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
  begin
    Icon := Items[i].Icon;
    if FileExists(Items[i].Icon) then
    begin
      Icon := 'file:///' + Icon;
      Icon := StringReplace(Icon, '\', '/', [rfReplaceAll]);
    end;

    Params := Format(StrParams, [
                    IntToStr(IdxList),
                    LowerCase(BoolToStr(Items[i].Animation.OnDrop, True)),
                    LowerCase(BoolToStr(Items[i].Animation.Bounce, True)),
                    LowerCase(BoolToStr(Items[i].Clickable, True)),
                    LowerCase(BoolToStr(Items[i].Draggable, True)),
                    LowerCase(BoolToStr(Items[i].Flat, True)),
                    LowerCase(BoolToStr(Items[i].Optimized, True)),
                    Items[i].Position.LatToStr(Map.Precision),
                    Items[i].Position.LngToStr(Map.Precision),
                    LowerCase(BoolToStr(Items[i].RaiseOnDrag, True)),
                    QuotedStr(Items[i].Title),
                    LowerCase(BoolToStr(Items[i].Visible, True)),
                    QuotedStr(Icon),
                    IntToStr(Items[i].Index),
                    QuotedStr(Items[i].InfoWindow.GetConvertedString),
                    LowerCase(BoolToStr(Items[i].InfoWindow.DisableAutoPan, True)),
                    IntToStr(Items[i].InfoWindow.MaxWidth),
                    IntToStr(Items[i].InfoWindow.PixelOffset.Height),
                    IntToStr(Items[i].InfoWindow.PixelOffset.Width),
                    LowerCase(BoolToStr(Items[i].InfoWindow.CloseOtherBeforeOpen, True))
                    ]);

    ExecuteScript('MakeMarker', Params);
  end;
end;

procedure TGMMarker.EventFired(EventType: TEventType; Params: array of const);
var
  LL: TLatLng;
begin
  inherited;

  if EventType = etInfoWinCloseClick then Exit;

  if High(Params) <> 2 then
    raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

  if (Params[0].VType <> vtExtended) or (Params[1].VType <> vtExtended) then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VType <> vtInteger then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VInteger > VisualObjects.Count - 1 then
    raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

  LL := TLatLng.Create(ControlPrecision(Params[0].VExtended^, Map.Precision),
                       ControlPrecision(Params[1].VExtended^, Map.Precision));

  case EventType of
    etMarkerClick: if Assigned(FOnClick) then FOnClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerDblClick: if Assigned(FOnDblClick) then FOnDblClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerDrag: if Assigned(FOnDrag) then FOnDrag(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerDragEnd:
    begin
      Items[Params[2].VInteger].Position.Assign(LL);
      if Assigned(FOnDragEnd) then FOnDragEnd(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    end;
    etMarkerDragStart: if Assigned(FOnDragStart) then FOnDragStart(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerMouseDown: if Assigned(FOnMouseDown) then FOnMouseDown(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerMouseOut: if Assigned(FOnMouseOut) then FOnMouseOut(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerMouseOver: if Assigned(FOnMouseOver) then FOnMouseOver(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerMouseUp: if Assigned(FOnMouseUp) then FOnMouseUp(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etMarkerRightClick: if Assigned(FOnRightClick) then FOnRightClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
  end;
end;

{ TMarkers }

function TMarkers.Add: TMarker;
begin
  Result := TMarker(inherited Add);
end;

function TMarkers.GetItems(I: Integer): TMarker;
begin
  Result := TMarker(inherited Items[I]);
end;

function TMarkers.GetOwner: TPersistent;
begin
  Result := TGMMarker(inherited GetOwner);
end;

function TMarkers.Insert(Index: Integer): TMarker;
begin
  Result := TMarker(inherited Insert(Index));
end;

procedure TMarkers.SetItems(I: Integer; const Value: TMarker);
begin
  inherited SetItem(I, Value);
end;

{ TAnimation }

procedure TAnimation.Assign(Source: TPersistent);
begin
  if Source is TAnimation then
  begin
    OnDrop := TAnimation(Source).OnDrop;
    Bounce := TAnimation(Source).Bounce;
  end
  else
    inherited Assign(Source);
end;

constructor TAnimation.Create(aOwner: TMarker);
begin
  inherited Create;

  FMarker := aOwner;
  FOnDrop := False;
  FBounce := False;

  if Assigned(FMarker) then FIdxList := FMarker.IdxList;
end;

procedure TAnimation.SetBounce(const Value: Boolean);
const
  StrParams = '%s,%s,%s';
var
  Params: String;
begin
  if FBounce = Value then Exit;

  FBounce := Value;

  if Assigned(FMarker) and (FMarker is TMarker) and
     Assigned(FMarker.Collection) and (FMarker.Collection is TMarkers) and
     Assigned(TMarkers(FMarker.Collection).FGMLinkedComponent) and
     TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).AutoUpdate then
  begin
    Params := Format(StrParams, [
                      IntToStr(FIdxList),
                      LowerCase(BoolToStr(FBounce, True)),
                      IntToStr(FMarker.Index)]);
    TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).ExecuteScript('MarkerSetBounce', Params);

    // control de errores // error control
    TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).ErrorControl;
  end;
end;

procedure TAnimation.SetOnDrop(const Value: Boolean);
begin
  if FOnDrop = Value then Exit;

  FOnDrop := Value;

  if Assigned(FMarker) and (FMarker is TMarker) and
     Assigned(FMarker.Collection) and (FMarker.Collection is TMarkers) and
     Assigned(TMarkers(FMarker.Collection).FGMLinkedComponent) and
     TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).AutoUpdate then
    TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).ShowElements;
end;

{ TMarker }

procedure TMarker.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TMarker then
  begin
    Animation.Assign(TMarker(Source).Animation);
    Clickable := TMarker(Source).Clickable;
    Draggable := TMarker(Source).Draggable;
    Flat := TMarker(Source).Flat;
    Optimized := TMarker(Source).Optimized;
    Position.Assign(TMarker(Source).Position);
    RaiseOnDrag := TMarker(Source).RaiseOnDrag;
    Title := TMarker(Source).Title;
    Visible := TMarker(Source).Visible;
  end;
end;

procedure TMarker.CenterMapToMarker;
begin
  if Assigned(Collection) and (Collection is TMarkers) and
     Assigned(TMarkers(Collection).FGMLinkedComponent) and
     Assigned(TMarkers(Collection).FGMLinkedComponent.Map) then
    TMarkers(Collection).FGMLinkedComponent.Map.SetCenter(Position.Lat, Position.Lng);
end;

constructor TMarker.Create(Collection: TCollection);
begin
  inherited;

  FAnimation := TAnimation.Create(Self);
  FClickable := True;
  FDraggable := False;
  FFlat := False;
  FOptimized := True;
  FPosition := TMarkerLatLng.Create(Self);
  FRaiseOnDrag := True;
  FTitle := '';
  FVisible := True;
end;

destructor TMarker.Destroy;
begin
  if Assigned(FPosition) then FreeAndNil(FPosition);
  if Assigned(FAnimation) then FreeAndNil(FAnimation);

  inherited;
end;

function TMarker.GetDisplayName: String;
begin
  if Length(Title) > 0 then
  begin
    if Length(Title) > 15 then
      Result := Copy(Title, 0, 12) + '...'
    else
      Result := Title;
  end
  else
    Result := inherited GetDisplayName;
end;

function TMarker.GetZIndex: Integer;
begin
  Result := Index;
end;

procedure TMarker.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;

  SetProperty('SetClickable', 'OnClickableChange', FClickable);
end;

procedure TMarker.SetDraggable(const Value: Boolean);
begin
  if FDraggable = Value then Exit;

  FDraggable := Value;

  SetProperty('MarkerSetDraggable', 'OnDragableChange', FDraggable);
end;

procedure TMarker.SetFlat(const Value: Boolean);
begin
  if FFlat = Value then Exit;

  FFlat := Value;

  SetProperty('MarkerSetFlat', 'OnFlatChange', FFlat);
end;

procedure TMarker.SetIcon(const Value: string);
begin
  if FIcon = Value then Exit;

  FIcon := Value;

  SetProperty('MarkerSetIcon', 'OnIconChange', FIcon);
end;

procedure TMarker.SetIdxList(const Value: Cardinal);
begin
  inherited;

  FAnimation.IdxList := Value;
  FPosition.IdxList := Value;
end;

procedure TMarker.SetOptimized(const Value: Boolean);
begin
  if FOptimized = Value then Exit;

  FOptimized := Value;

  if Assigned(Collection) and (Collection is TMarkers) and
     Assigned(TMarkers(Collection).FGMLinkedComponent) and
     TGMMarker(TMarkers(Collection).FGMLinkedComponent).AutoUpdate then
    TGMMarker(TMarkers(Collection).FGMLinkedComponent).ShowElements;
end;

procedure TMarker.SetRaiseOnDrag(const Value: Boolean);
begin
  if FRaiseOnDrag = Value then Exit;

  FRaiseOnDrag := Value;

  if Assigned(Collection) and (Collection is TMarkers) and
     Assigned(TMarkers(Collection).FGMLinkedComponent) and
     TGMMarker(TMarkers(Collection).FGMLinkedComponent).AutoUpdate then
    TGMMarker(TMarkers(Collection).FGMLinkedComponent).ShowElements;
end;

procedure TMarker.SetTitle(const Value: string);
begin
  if FTitle = Value then Exit;

  FTitle := Value;

  SetProperty('MarkerSetTitle', 'OnTitleChange', FTitle);
end;

procedure TMarker.SetVisible(const Value: Boolean);
begin
  if FVisible = Value then Exit;

  FVisible := Value;

  SetProperty('SetVisible', 'OnVisibleChange', FVisible);
end;

{ TMarkerLatLng }

procedure TMarkerLatLng.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TMarkerLatLng then
    IdxList := TMarkerLatLng(Source).IdxList;
end;

constructor TMarkerLatLng.Create(Marker: TMarker; Lat, Lng: Real);
begin
  inherited Create(Lat, Lng);

  FMarker := Marker;
  if Assigned(FMarker) then FIdxList := FMarker.IdxList;
end;

procedure TMarkerLatLng.PositionChange;
const
  StrParams = '%s,%s,%s,%s';
var
  Params: String;
begin
  if Assigned(FMarker) and (FMarker is TMarker) and
     Assigned(FMarker.Collection) and (FMarker.Collection is TMarkers) and
     Assigned(TMarkers(FMarker.Collection).FGMLinkedComponent) and
     TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).AutoUpdate and
     Assigned(TMarkers(FMarker.Collection).FGMLinkedComponent.Map) then
  begin
    Params := Format(StrParams, [
                   IntToStr(FIdxList),
                   FMarker.Position.LatToStr(TMarkers(FMarker.Collection).FGMLinkedComponent.Map.Precision),
                   FMarker.Position.LngToStr(TMarkers(FMarker.Collection).FGMLinkedComponent.Map.Precision),
                   IntToStr(FMarker.Index)
                  ]);
    TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).ExecuteScript('MarkerSetPosition', Params);

    // ES/EN: evento/event
    if Assigned(TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).OnPositionChange) then
      TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent).OnPositionChange(TGMMarker(TMarkers(FMarker.Collection).FGMLinkedComponent), Self, FMarker.Index, FMarker);
  end;
end;

procedure TMarkerLatLng.SetLat(const Value: Real);
begin
  if Lat = Value then Exit;

  inherited;

  PositionChange;
end;

procedure TMarkerLatLng.SetLng(const Value: Real);
begin
  if Lng = Value then Exit;

  inherited;

  PositionChange;
end;

end.
