{
TGMPolyline component

  ES: componente para mostrar polilineas en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to show polylines on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner las
      polilineas a mostrar
  EN: put the component into a form, link to a TGMMap and put the polylines to show
=========================================================================
History:

ver 0.1.4
  ES:
    cambio: cambio en el JavaScript de algunos mtodos Set
  EN:
    change: JavaScript changed from some Set methods

ver 0.1.3
  ES:
    cambio: mtodos Set cambiados para evitar duplicidad de cdigo
  EN:
    change: changed Set methods to avoid duplicate code

ver 0.1.2
  ES:
    cambio: cuando cambia la lat/lng de un TLinePoint, se actualiza en el mapa
    error: implementado SetPath en JavaScript
  EN:
    change: when change the lat/lng of a TLinePoint, the map is updated
    bug: implemented SetPath into JavaScript

ver 0.1.1:
  ES- primera versin
  EN- first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMPolyline;

interface

uses
  Classes, Graphics,
  GMMap, GMLatLng;

type
  TPolyline = class;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TLinePoint = class(TCollectionItem)
  private
    FLatLng: TLatLng;
    function GetLat: Real;
    function GetLng: Real;
  protected
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure SetLat(const Value: Real);
    procedure SetLng(const Value: Real);

    function LatToStr(Precision: Integer = 6): string;
    function LngToStr(Precision: Integer = 6): string;
  published
    property Lat: Real read GetLat write SetLat;
    property Lng: Real read GetLng write SetLng;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TLinePoints = class(TCollection)
  private
    function GetItems(I: Integer): TLinePoint;
    procedure SetItems(I: Integer; const Value: TLinePoint);
  protected
    FPolyline: TPolyline;

    function GetOwner: TPersistent; override;

    function Add: TLinePoint;
    function Insert(Index: Integer): TLinePoint;
    procedure Delete(Index: Integer);
    procedure Move(CurIndex, NewIndex: Integer);
    procedure Clear;

    property Items[I: Integer]: TLinePoint read GetItems write SetItems; default;
  public
    constructor Create(Polyline: TPolyline; ItemClass: TCollectionItemClass);

    procedure Assign(Source: TPersistent); override;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TPolyline = class(TLinkedComponent)
  private
    FStrokeColor: TColor;
    FVisible: Boolean;
    FStrokeWeight: Integer;
    FGeodesic: Boolean;
    FClickable: Boolean;
    FStrokeOpacity: Real;
    FEditable: Boolean;
    FLinePoints: TLinePoints;
    FAutoUpdatePath: Boolean;
    procedure SetClickable(const Value: Boolean);
    procedure SetEditable(const Value: Boolean);
    procedure SetGeodesic(const Value: Boolean);
    procedure SetStrokeColor(const Value: TColor);
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
    procedure SetVisible(const Value: Boolean);
    function GetItems(I: Integer): TLinePoint;
  protected
    procedure SetIdxList(const Value: Cardinal); override;

    function PolylineToStr: string;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure CenterMapTo;
    procedure GetPath;
    procedure SetPath;

    function AddLinePoint(Lat, Lng: Real): TLinePoint; overload;
    function AddLinePoint(Lat, Lng: string): TLinePoint; overload;
    function InsertLinePoint(Index: Integer; Lat, Lng: Real): TLinePoint;
    procedure DeleteLinePoint(Index: Integer);
    procedure MoveLinePoint(CurIndex, NewIndex: Integer);
    procedure ClearLinePoints;

    property Items[I: Integer]: TLinePoint read GetItems; default;
  published
    property Clickable: Boolean read FClickable write SetClickable;
    property Editable: Boolean read FEditable write SetEditable;
    property Geodesic: Boolean read FGeodesic write SetGeodesic;
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
    property Visible: Boolean read FVisible write SetVisible;
    property AutoUpdatePath: Boolean read FAutoUpdatePath write FAutoUpdatePath;
    // ES/EN: TCollection
    property LinePoints: TLinePoints read FLinePoints write FLinePoints;
    property InfoWindow;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TPolylines = class(TLinkedComponents)
  private
    procedure SetItems(I: Integer; const Value: TPolyline);
    function GetItems(I: Integer): TPolyline;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TPolyline;
    function Insert(Index: Integer): TPolyline;

    property Items[I: Integer]: TPolyline read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Polyline
  **************************************************************************** }
  TGMPolyline = class(TGMLinkedComponent)
  private
    FOnRightClick: TLatLngIdxEvent;
    FOnMouseDown: TLatLngIdxEvent;
    FOnMouseMove: TLatLngIdxEvent;
    FOnMouseUp: TLatLngIdxEvent;
    FOnMouseOut: TLatLngIdxEvent;
    FOnDblClick: TLatLngIdxEvent;
    FOnMouseOver: TLatLngIdxEvent;
    FOnClick: TLatLngIdxEvent;
    FOnStrokeColorChange: TLinkedComponentChange;
    FOnVisibleChange: TLinkedComponentChange;
    FOnStrokeWeightChange: TLinkedComponentChange;
    FOnGeodesicChange: TLinkedComponentChange;
    FOnClickableChange: TLinkedComponentChange;
    FOnStrokeOpacityChange: TLinkedComponentChange;
    FOnEditableChange: TLinkedComponentChange;
    FOnPathChange: TLinkedComponentChange;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TPolyline;

    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add: TPolyline;
    procedure GetPath;
    procedure SetPath;

    property Items[I: Integer]: TPolyline read GetItems; default;
  published
    property VisualObjects;
    // eventos
    // events
        // from map
    property OnClick: TLatLngIdxEvent read FOnClick write FOnClick;
    property OnDblClick: TLatLngIdxEvent read FOnDblClick write FOnDblClick;
    property OnMouseDown: TLatLngIdxEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TLatLngIdxEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseOut: TLatLngIdxEvent read FOnMouseOut write FOnMouseOut;
    property OnMouseOver: TLatLngIdxEvent read FOnMouseOver write FOnMouseOver;
    property OnMouseUp: TLatLngIdxEvent read FOnMouseUp write FOnMouseUp;
    property OnRightClick: TLatLngIdxEvent read FOnRightClick write FOnRightClick;
        // from properties
    property OnClickableChange: TLinkedComponentChange read FOnClickableChange write FOnClickableChange;
    property OnEditableChange: TLinkedComponentChange read FOnEditableChange write FOnEditableChange;
    property OnGeodesicChange: TLinkedComponentChange read FOnGeodesicChange write FOnGeodesicChange;
    property OnStrokeColorChange: TLinkedComponentChange read FOnStrokeColorChange write FOnStrokeColorChange;
    property OnStrokeOpacityChange: TLinkedComponentChange read FOnStrokeOpacityChange write FOnStrokeOpacityChange;
    property OnStrokeWeightChange: TLinkedComponentChange read FOnStrokeWeightChange write FOnStrokeWeightChange;
    property OnVisibleChange: TLinkedComponentChange read FOnVisibleChange write FOnVisibleChange;
    property OnPathChange: TLinkedComponentChange read FOnPathChange write FOnPathChange;
  end;

implementation

uses
  SysUtils,
  Lang, WebControl, GMConstants;

{ TGMPolyline }

function TGMPolyline.Add: TPolyline;
begin
  Result := TPolyline(inherited Add);
end;

procedure TGMPolyline.EventFired(EventType: TEventType; Params: array of const);
var
  LL: TLatLng;
begin
  inherited;

  if EventType = etInfoWinCloseClick then Exit;

  if High(Params) <> 2 then
    raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

  if (Params[0].VType <> vtExtended) or (Params[1].VType <> vtExtended) then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VType <> vtInteger then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VInteger > VisualObjects.Count - 1 then
    raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

  LL := TLatLng.Create(ControlPrecision(Params[0].VExtended^, Map.Precision),
                       ControlPrecision(Params[1].VExtended^, Map.Precision));

  case EventType of
    etPolylineClick: if Assigned(FOnClick) then FOnClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineDblClick: if Assigned(FOnDblClick) then FOnDblClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineMouseDown: if Assigned(FOnMouseDown) then FOnMouseDown(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineMouseMove: if Assigned(FOnMouseMove) then FOnMouseMove(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineMouseOut: if Assigned(FOnMouseOut) then FOnMouseOut(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineMouseOver: if Assigned(FOnMouseOver) then FOnMouseOver(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineMouseUp: if Assigned(FOnMouseUp) then FOnMouseUp(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etPolylineRightClick: if Assigned(FOnRightClick) then FOnRightClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
  end;
end;

function TGMPolyline.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Polyline';
end;

function TGMPolyline.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TPolylines;
end;

function TGMPolyline.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TPolyline;
end;

function TGMPolyline.GetItems(I: Integer): TPolyline;
begin
  Result := TPolyline(inherited Items[i]);
end;

procedure TGMPolyline.GetPath;
var
  i: Integer;
begin
  if not Assigned(Map) then Exit;

  for i := 0 to VisualObjects.Count - 1 do
    Items[i].GetPath;
end;

procedure TGMPolyline.SetPath;
var
  i: Integer;
begin
  if not Assigned(Map) then Exit;

  for i := 0 to VisualObjects.Count - 1 do
    Items[i].SetPath;
end;

procedure TGMPolyline.ShowElements;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
  i: Integer;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
  begin
    Params := Format(StrParams, [
                    IntToStr(IdxList),
                    LowerCase(BoolToStr(Items[i].Clickable, True)),
                    LowerCase(BoolToStr(Items[i].Editable, True)),
                    LowerCase(BoolToStr(Items[i].Geodesic, True)),
                    QuotedStr(Map.TColorToStr(Items[i].StrokeColor)),
                    StringReplace(FloatToStr(Items[i].StrokeOpacity), ',', '.', [rfReplaceAll]),
                    IntToStr(Items[i].StrokeWeight),
                    LowerCase(BoolToStr(Items[i].Visible, True)),
                    QuotedStr(Items[i].PolylineToStr),
                    IntToStr(Items[i].Index),
                    QuotedStr(Items[i].InfoWindow.GetConvertedString),
                    LowerCase(BoolToStr(Items[i].InfoWindow.DisableAutoPan, True)),
                    IntToStr(Items[i].InfoWindow.MaxWidth),
                    IntToStr(Items[i].InfoWindow.PixelOffset.Height),
                    IntToStr(Items[i].InfoWindow.PixelOffset.Width),
                    LowerCase(BoolToStr(Items[i].InfoWindow.CloseOtherBeforeOpen, True))
                    ]);

    ExecuteScript('MakePolyline', Params);
  end;
end;

{ TPolylines }

function TPolylines.Add: TPolyline;
begin
  Result := TPolyline(inherited Add);
end;

function TPolylines.GetItems(I: Integer): TPolyline;
begin
  Result := TPolyline(inherited Items[I]);
end;

function TPolylines.GetOwner: TPersistent;
begin
  Result := TGMPolyline(inherited GetOwner);
end;

function TPolylines.Insert(Index: Integer): TPolyline;
begin
  Result := TPolyline(inherited Insert(Index));
end;

procedure TPolylines.SetItems(I: Integer; const Value: TPolyline);
begin
  inherited SetItem(I, Value);
end;

{ TPolyline }

function TPolyline.AddLinePoint(Lat, Lng: Real): TLinePoint;
begin
  Result := LinePoints.Add;
  Result.Lat := Lat;
  Result.Lng := Lng;
end;

function TPolyline.AddLinePoint(Lat, Lng: string): TLinePoint;
begin
  Result := LinePoints.Add;
  Result.Lat := Result.FLatLng.StringToReal(Lat);
  Result.Lng := Result.FLatLng.StringToReal(Lng);
end;

procedure TPolyline.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TPolyline then
  begin
    Clickable := TPolyline(Source).Clickable;
    Editable  := TPolyline(Source).Editable;
    Geodesic := TPolyline(Source).Geodesic;
    StrokeColor := TPolyline(Source).StrokeColor;
    StrokeOpacity := TPolyline(Source).StrokeOpacity;
    StrokeWeight := TPolyline(Source).StrokeWeight;
    Visible := TPolyline(Source).Visible;
    LinePoints.Assign(TPolyline(Source).LinePoints);
    AutoUpdatePath := TPolyline(Source).AutoUpdatePath;
  end;
end;

procedure TPolyline.CenterMapTo;
begin
  if Assigned(Collection) and (Collection is TPolylines) and
     Assigned(TPolylines(Collection).FGMLinkedComponent) and
     Assigned(TPolylines(Collection).FGMLinkedComponent.Map) and
     (LinePoints.Count > 0) then
    TPolylines(Collection).FGMLinkedComponent.Map.SetCenter(Items[0].Lat, Items[0].Lat);
end;

procedure TPolyline.ClearLinePoints;
begin
  LinePoints.Clear;
end;

constructor TPolyline.Create(Collection: TCollection);
begin
  inherited;

  FClickable := True;
  FEditable := False;
  FGeodesic := False;
  FStrokeColor := clBlack;
  FStrokeOpacity := 1;
  FStrokeWeight := 2;
  FVisible := True;
  FLinePoints := TLinePoints.Create(Self, TLinePoint);
  FAutoUpdatePath := True;
end;

procedure TPolyline.DeleteLinePoint(Index: Integer);
begin
  LinePoints.Delete(Index);
end;

destructor TPolyline.Destroy;
begin
  if Assigned(FLinePoints) then FreeAndNil(FLinePoints);

  inherited;
end;

function TPolyline.GetItems(I: Integer): TLinePoint;
begin
  Result := TLinePoint(FLinePoints[i]);
end;

procedure TPolyline.GetPath;
const
  StrParams = '%s,%s';
var
  WC: TWebControl;
  Params: string;
  Path: string;
  i: Integer;
  L1, L2: TStringList;
  IsEqual: Boolean;
  LL: TLatLng;
  AutoUpdate: Boolean;
begin
  if not Assigned(Collection) or not (Collection is TPolylines) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent.Map) then
    Exit;

  Params := Format(StrParams, [IntToStr(IdxList), IntToStr(Index)]);
  if not TGMPolyline(TPolylines(Collection).FGMLinkedComponent).ExecuteScript('PolylineGetPath', Params) then
    Exit;

  WC := TWebControl.Create(TPolylines(Collection).FGMLinkedComponent.Map.WebBrowser);
  try
    Path := WC.GetStringField(PolylineForm, PolylineFormPath);
  finally
    FreeAndNil(WC);
  end;

  if Path = '' then Exit;

  L1 := TStringList.Create;
  L2 := TStringList.Create;
  LL := TLatLng.Create;
  AutoUpdate := TPolylines(Collection).FGMLinkedComponent.AutoUpdate;
  try
    L1.Delimiter := ';';
    L2.Delimiter := '|';
    L1.DelimitedText := Path;

    // ES: miramos si no ha cambiado // EN: check if not have changed
    if LinePoints.Count = L1.Count then
    begin
      IsEqual := True;
      for i := 0 to L1.Count - 1 do
      begin
        L2.DelimitedText := L1[i];
        LL.Lat := LL.StringToReal(L2[0]);
        LL.Lng := LL.StringToReal(L2[1]);
        if not LL.IsEqual(Items[i].FLatLng) then
        begin
          IsEqual := False;
          Break;
        end;
      end;

      if IsEqual then
      begin
        if Assigned(LL) then FreeAndNil(LL);
        if Assigned(L1) then FreeAndNil(L1);
        if Assigned(L2) then FreeAndNil(L2);
        Exit;
      end;
    end;

    TPolylines(Collection).FGMLinkedComponent.AutoUpdate := False;
    LinePoints.Clear;
    for i := 0 to L1.Count - 1 do
    begin
      L2.DelimitedText := L1[i];
      if L2.Count <> 2 then Break;
      AddLinePoint(L2[0], L2[1]);
    end;
  finally
    TPolylines(Collection).FGMLinkedComponent.AutoUpdate := AutoUpdate;
    if Assigned(LL) then FreeAndNil(LL);
    if Assigned(L1) then FreeAndNil(L1);
    if Assigned(L2) then FreeAndNil(L2);
  end;

  if Assigned(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).OnPathChange) then
    TGMPolyline(TPolylines(Collection).FGMLinkedComponent).OnPathChange(TGMPolyline(TPolylines(Collection).FGMLinkedComponent), Index, Self);
end;

function TPolyline.InsertLinePoint(Index: Integer; Lat, Lng: Real): TLinePoint;
begin
  Result := LinePoints.Insert(Index);
  Result.Lat := Lat;
  Result.Lng := Lng;
end;

procedure TPolyline.MoveLinePoint(CurIndex, NewIndex: Integer);
begin
  LinePoints.Move(CurIndex, NewIndex);
end;

function TPolyline.PolylineToStr: string;
var
  i: Integer;
  Prec: Integer;
begin
  Result := '';

  if not Assigned(Collection) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent) or
     not Assigned(TPolylines(Collection).FGMLinkedComponent.Map) then
    Exit;

  Prec := TPolylines(Collection).FGMLinkedComponent.Map.Precision;
  for i := 0 to FLinePoints.Count - 1 do
    if Result <> '' then
      Result := Result + ';' + FLinePoints[i].LatToStr(Prec) + '|' + FLinePoints[i].LngToStr(Prec)
    else
      Result := FLinePoints[i].LatToStr(Prec) + '|' + FLinePoints[i].LngToStr(Prec);
end;

procedure TPolyline.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;

  SetProperty('SetClickable', 'OnClickableChange', FClickable);
end;

procedure TPolyline.SetEditable(const Value: Boolean);
begin
  if FEditable = Value then Exit;

  FEditable := Value;

  SetProperty('SetEditable', 'OnEditableChange', FEditable);
end;

procedure TPolyline.SetGeodesic(const Value: Boolean);
begin
  if FGeodesic = Value then Exit;

  FGeodesic := Value;

  SetProperty('PolylineSetGeodesic', 'OnGeodesicChange', FGeodesic);
end;

procedure TPolyline.SetIdxList(const Value: Cardinal);
begin
  inherited;

end;

procedure TPolyline.SetPath;
const
  StrParams = '%s,%s,%s';
var
  Params: string;
begin
  if Assigned(Collection) and (Collection is TPolylines) and
     Assigned(TPolylines(Collection).FGMLinkedComponent) and
     TGMPolyline(TPolylines(Collection).FGMLinkedComponent).AutoUpdate then
  begin
    Params := Format(StrParams, [
                      IntToStr(IdxList),
                      QuotedStr(PolylineToStr),
                      IntToStr(Index)]);
    TGMPolyline(TPolylines(Collection).FGMLinkedComponent).ExecuteScript('PolylineSetPath', Params);

    // ES: control de errores // EN: error control
    TGMPolyline(TPolylines(Collection).FGMLinkedComponent).ErrorControl;

    // ES/EN: evento/event
    if Assigned(TGMPolyline(TPolylines(Collection).FGMLinkedComponent).OnStrokeColorChange) then
      TGMPolyline(TPolylines(Collection).FGMLinkedComponent).OnStrokeColorChange(TGMPolyline(TPolylines(Collection).FGMLinkedComponent), Index, Self);
  end;
end;

procedure TPolyline.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;

  {$IF CompilerVersion < 20}  // ES: si la versin es inferior a D2009 - EN: if lower than D2009 version
  SetColorProperty('SetStrokeColor', 'OnStrokeColorChange', FStrokeColor);
  {$ELSE}
  SetProperty('SetStrokeColor', 'OnStrokeColorChange', FStrokeColor);
  {$IFEND}
end;

procedure TPolyline.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;
  if FStrokeOpacity < 0 then FStrokeOpacity := 0;
  if FStrokeOpacity > 1 then FStrokeOpacity := 1;
  FStrokeOpacity := Trunc(FStrokeOpacity * 100) / 100;

  SetProperty('SetStrokeOpacity', 'OnStrokeOpacityChange', FStrokeOpacity);
end;

procedure TPolyline.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;
  if FStrokeWeight < 1 then FStrokeWeight := 1;
  if FStrokeWeight > 10 then FStrokeWeight := 10;

  SetProperty('SetStrokeWeight', 'OnStrokeWeightChange', FStrokeWeight);
end;

procedure TPolyline.SetVisible(const Value: Boolean);
begin
  if FVisible = Value then Exit;

  FVisible := Value;

  SetProperty('SetVisible', 'OnVisibleChange', FVisible);
end;

{ TLinePoints }

function TLinePoints.Add: TLinePoint;
begin
  Result := TLinePoint(inherited Add);

  if Assigned(FPolyline) and Assigned(FPolyline.Collection) and
     Assigned(TPolylines(FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    FPolyline.SetPath;
end;

procedure TLinePoints.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TLinePoints then
    FPolyline := TLinePoints(Source).FPolyline;
end;

procedure TLinePoints.Clear;
begin
  inherited Clear;

  if Assigned(FPolyline) and Assigned(FPolyline.Collection) and
     Assigned(TPolylines(FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    TGMPolyline(TPolylines(FPolyline.Collection).FGMLinkedComponent).ShowElements;
end;

constructor TLinePoints.Create(Polyline: TPolyline; ItemClass: TCollectionItemClass);
begin
  inherited Create(ItemClass);

  FPolyline := Polyline;
end;

procedure TLinePoints.Delete(Index: Integer);
begin
  inherited Delete(Index);

  if Assigned(FPolyline) and Assigned(FPolyline.Collection) and
     Assigned(TPolylines(FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    FPolyline.SetPath;
end;

function TLinePoints.GetItems(I: Integer): TLinePoint;
begin
  Result := TLinePoint(inherited Items[I]);
end;

function TLinePoints.GetOwner: TPersistent;
begin
  Result := FPolyline;
end;

function TLinePoints.Insert(Index: Integer): TLinePoint;
begin
  Result := TLinePoint(inherited Insert(Index));

  if Assigned(FPolyline) and Assigned(FPolyline.Collection) and
     Assigned(TPolylines(FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    FPolyline.SetPath;
end;

procedure TLinePoints.Move(CurIndex, NewIndex: Integer);
begin
  Items[CurIndex].Index := NewIndex;

  if Assigned(FPolyline) and Assigned(FPolyline.Collection) and
     Assigned(TPolylines(FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    FPolyline.SetPath;
end;

procedure TLinePoints.SetItems(I: Integer; const Value: TLinePoint);
begin
  inherited SetItem(I, Value);
end;

{ TLinePoint }

procedure TLinePoint.Assign(Source: TPersistent);
begin
  if Source is TLinePoint then
  begin
    Lat := TLinePoint(Source).Lat;
    Lng := TLinePoint(Source).Lng;
  end
  else
    inherited Assign(Source);
end;

constructor TLinePoint.Create(Collection: TCollection);
begin
  inherited;

  FLatLng := TLatLng.Create;
end;

destructor TLinePoint.Destroy;
begin
  if Assigned(FLatLng) then FreeAndNil(FLatLng);

  inherited;
end;

function TLinePoint.GetLat: Real;
begin
  Result := FLatLng.Lat;
end;

function TLinePoint.GetLng: Real;
begin
  Result := FLatLng.Lng;
end;

function TLinePoint.LatToStr(Precision: Integer): string;
begin
  Result := FLatLng.LatToStr(Precision);
end;

function TLinePoint.LngToStr(Precision: Integer): string;
begin
  Result := FLatLng.LngToStr(Precision);
end;

procedure TLinePoint.SetLat(const Value: Real);
begin
  if FLatLng.Lat = Value then Exit;

  FLatLng.Lat := Value;

  if Assigned(Collection) and
     Assigned(TLinePoints(Collection).FPolyline) and
     Assigned(TLinePoints(Collection).FPolyline.Collection) and
     Assigned(TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    TLinePoints(Collection).FPolyline.SetPath;
end;

procedure TLinePoint.SetLng(const Value: Real);
begin
  if FLatLng.Lng = Value then Exit;

  FLatLng.Lng := Value;

  if Assigned(Collection) and
     Assigned(TLinePoints(Collection).FPolyline) and
     Assigned(TLinePoints(Collection).FPolyline.Collection) and
     Assigned(TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent) and
     TPolylines(TLinePoints(Collection).FPolyline.Collection).FGMLinkedComponent.AutoUpdate then
    TLinePoints(Collection).FPolyline.SetPath;
end;

end.
