{
TGMRectangle component

  ES: componente para crear rectngulos en un mapa de Google Maps mediante el
      componente TGMMap
  EN: component to create rectangles on Google Map map using the component TGMMap

=========================================================================
MODO DE USO/HOW TO USE

  ES: poner el componente en el formulario, linkarlo a un TGMMap y poner los
      rectngulos a mostrar
  EN: put the component into a form, link to a TGMMap and put the rectangles to
      show
=========================================================================
History:

ver 0.1.4
  ES:
    cambio: cambio en el JavaScript de algunos mtodos Set
  EN:
    change: JavaScript changed from some Set methods

ver 0.1.3
  ES: primera versin
  EN: first version
=========================================================================
IMPORTANTE PROGRAMADORES: Por favor, si tienes comentarios, mejoras,
  ampliaciones, errores y/o cualquier otro tipo de sugerencia, envame un correo a:
  gmlib@cadetill.com

IMPORTANT PROGRAMMERS: please, if you have comments, improvements, enlargements,
  errors and/or any another type of suggestion, please send me a mail to:
  gmlib@cadetill.com
=========================================================================

Copyright () 2012, by Xavier Martinez (cadetill)

@author Xavier Martinez (cadetill)
@web  http://www.cadetill.com
}
unit GMRectangle;

interface

uses
  Classes, Graphics,
  GMMap, GMLatLngBounds;

type
  { ****************************************************************************
    internal class
  **************************************************************************** }
  TRectangle = class(TLinkedComponent)
  private
    FFillColor: TColor;
    FStrokeColor: TColor;
    FVisible: Boolean;
    FBounds: TLatLngBounds;
    FStrokeWeight: Integer;
    FFillOpacity: Real;
    FClickable: Boolean;
    FStrokeOpacity: Real;
    FEditable: Boolean;
    FIsBoundsUpdt: Boolean;
    procedure SetClickable(const Value: Boolean);
    procedure SetEditable(const Value: Boolean);
    procedure SetFillColor(const Value: TColor);
    procedure SetFillOpacity(const Value: Real);
    procedure SetStrokeColor(const Value: TColor);
    procedure SetStrokeOpacity(const Value: Real);
    procedure SetStrokeWeight(const Value: Integer);
    procedure SetVisible(const Value: Boolean);
    procedure OnChangeBounds(Sender: TObject);
  protected
    procedure SetIdxList(const Value: Cardinal); override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;

    procedure CenterMapTo;
  published
    property Bounds: TLatLngBounds read FBounds write FBounds;
    property Clickable: Boolean read FClickable write SetClickable;
    property Editable: Boolean read FEditable write SetEditable;
    property FillColor: TColor read FFillColor write SetFillColor;
    property FillOpacity: Real read FFillOpacity write SetFillOpacity; // 0 to 1
    property StrokeColor: TColor read FStrokeColor write SetStrokeColor;
    property StrokeOpacity: Real read FStrokeOpacity write SetStrokeOpacity; // 0 to 1
    property StrokeWeight: Integer read FStrokeWeight write SetStrokeWeight; // 1 to 10
    property Visible: Boolean read FVisible write SetVisible;
    property InfoWindow;
  end;

  { ****************************************************************************
    internal class
  **************************************************************************** }
  TRectangles = class(TLinkedComponents)
  private
    procedure SetItems(I: Integer; const Value: TRectangle);
    function GetItems(I: Integer): TRectangle;
  protected
    function GetOwner: TPersistent; override;
  public
    function Add: TRectangle;
    function Insert(Index: Integer): TRectangle;

    property Items[I: Integer]: TRectangle read GetItems write SetItems; default;
  end;

  { ****************************************************************************
    https://developers.google.com/maps/documentation/javascript/reference?hl=en#Rectangle
  **************************************************************************** }
  TGMRectangle = class(TGMLinkedComponent)
  private
    FOnStrokeColorChange: TLinkedComponentChange;
    FOnRightClick: TLatLngIdxEvent;
    FOnMouseDown: TLatLngIdxEvent;
    FOnVisibleChange: TLinkedComponentChange;
    FOnMouseMove: TLatLngIdxEvent;
    FOnStrokeWeightChange: TLinkedComponentChange;
    FOnMouseUp: TLatLngIdxEvent;
    FOnFillOpacityChange: TLinkedComponentChange;
    FOnClickableChange: TLinkedComponentChange;
    FOnMouseOut: TLatLngIdxEvent;
    FOnStrokeOpacityChange: TLinkedComponentChange;
    FOnEditableChange: TLinkedComponentChange;
    FOnBoundsChanged: TLinkedComponentChange;
    FOnDblClick: TLatLngIdxEvent;
    FOnMouseOver: TLatLngIdxEvent;
    FOnFillColorChange: TLinkedComponentChange;
    FOnClick: TLatLngIdxEvent;
  protected
    function GetAPIUrl: string; override;

    function GetItems(I: Integer): TRectangle;

    procedure ShowElements; override;
    procedure EventFired(EventType: TEventType; Params: array of const); override;
    function GetCollectionItemClass: TLinkedComponentClass; override;
    function GetCollectionClass: TLinkedComponentsClass; override;
  public
    function Add(SWLat: Real = 0; SWLng: Real = 0; NELat: Real = 0; NELng: Real = 0): TRectangle;

    property Items[I: Integer]: TRectangle read GetItems; default;
  published
    property VisualObjects;
    // eventos
    // events
        // both (map and properties)
    property OnBoundsChanged: TLinkedComponentChange read FOnBoundsChanged write FOnBoundsChanged;
        // from map
    property OnClick: TLatLngIdxEvent read FOnClick write FOnClick;
    property OnDblClick: TLatLngIdxEvent read FOnDblClick write FOnDblClick;
    property OnMouseDown: TLatLngIdxEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove: TLatLngIdxEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseOut: TLatLngIdxEvent read FOnMouseOut write FOnMouseOut;
    property OnMouseOver: TLatLngIdxEvent read FOnMouseOver write FOnMouseOver;
    property OnMouseUp: TLatLngIdxEvent read FOnMouseUp write FOnMouseUp;
    property OnRightClick: TLatLngIdxEvent read FOnRightClick write FOnRightClick;
        // from properties
    property OnClickableChange: TLinkedComponentChange read FOnClickableChange write FOnClickableChange;
    property OnEditableChange: TLinkedComponentChange read FOnEditableChange write FOnEditableChange;
    property OnFillColorChange: TLinkedComponentChange read FOnFillColorChange write FOnFillColorChange;
    property OnFillOpacityChange: TLinkedComponentChange read FOnFillOpacityChange write FOnFillOpacityChange;
    property OnStrokeColorChange: TLinkedComponentChange read FOnStrokeColorChange write FOnStrokeColorChange;
    property OnStrokeOpacityChange: TLinkedComponentChange read FOnStrokeOpacityChange write FOnStrokeOpacityChange;
    property OnStrokeWeightChange: TLinkedComponentChange read FOnStrokeWeightChange write FOnStrokeWeightChange;
    property OnVisibleChange: TLinkedComponentChange read FOnVisibleChange write FOnVisibleChange;
  end;

implementation

uses
  GMLatLng, Lang,
  SysUtils;

{ TGMRectangle }

function TGMRectangle.Add(SWLat, SWLng, NELat, NELng: Real): TRectangle;
begin
  Result := TRectangle(inherited Add);
  Result.Bounds.SW.Lat := SWLat;
  Result.Bounds.SW.Lng := SWLng;
  Result.Bounds.NE.Lat := NELat;
  Result.Bounds.NE.Lng := NELng;
end;

procedure TGMRectangle.EventFired(EventType: TEventType;
  Params: array of const);
var
  LL: TLatLng;
  OldEvent: TNotifyEvent;
begin
  inherited;

  if EventType = etInfoWinCloseClick then Exit;

  if EventType = etRectangleBoundsChange then
  begin
    if High(Params) <> 4 then
      raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

    if (Params[1].VType <> vtExtended) or (Params[2].VType <> vtExtended) or
       (Params[3].VType <> vtExtended) or (Params[4].VType <> vtExtended) then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[0].VType <> vtInteger then
      raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

    if Params[0].VInteger > VisualObjects.Count - 1 then
      raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

    if Items[Params[0].VInteger].FIsBoundsUpdt then
    begin
      Items[Params[0].VInteger].FIsBoundsUpdt := False;
      Exit;
    end;

    OldEvent := Items[Params[0].VInteger].Bounds.NE.OnChange;
    Items[Params[0].VInteger].Bounds.NE.OnChange := nil;
    Items[Params[0].VInteger].Bounds.SW.OnChange := nil;

    Items[Params[0].VInteger].Bounds.NE.Lat := ControlPrecision(Params[1].VExtended^, Map.Precision);
    Items[Params[0].VInteger].Bounds.NE.Lng := ControlPrecision(Params[2].VExtended^, Map.Precision);
    Items[Params[0].VInteger].Bounds.SW.Lat := ControlPrecision(Params[3].VExtended^, Map.Precision);
    Items[Params[0].VInteger].Bounds.SW.Lng := ControlPrecision(Params[4].VExtended^, Map.Precision);
    if Assigned(FOnBoundsChanged) then FOnBoundsChanged(Self, Params[0].VInteger, Items[Params[0].VInteger]);

    Items[Params[0].VInteger].Bounds.NE.OnChange := OldEvent;
    Items[Params[0].VInteger].Bounds.SW.OnChange := OldEvent;

    Exit;
  end;

  if High(Params) <> 2 then
    raise Exception.Create(GetTranslateText('Nmero de parmetros incorrecto', Map.Language));

  if (Params[0].VType <> vtExtended) or (Params[1].VType <> vtExtended) then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VType <> vtInteger then
    raise Exception.Create(GetTranslateText('Tipo de parmetro incorrecto', Map.Language));

  if Params[2].VInteger > VisualObjects.Count - 1 then
    raise Exception.Create(GetTranslateText('Valor de parmetro incorrecto', Map.Language));

  LL := TLatLng.Create(ControlPrecision(Params[0].VExtended^, Map.Precision),
                       ControlPrecision(Params[1].VExtended^, Map.Precision));

  case EventType of
    etRectangleClick: if Assigned(FOnClick) then FOnClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleDblClick: if Assigned(FOnDblClick) then FOnDblClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleMouseDown: if Assigned(FOnMouseDown) then FOnMouseDown(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleMouseMove: if Assigned(FOnMouseMove) then FOnMouseMove(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleMouseOut: if Assigned(FOnMouseOut) then FOnMouseOut(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleMouseOver: if Assigned(FOnMouseOver) then FOnMouseOver(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleMouseUp: if Assigned(FOnMouseUp) then FOnMouseUp(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
    etRectangleRightClick: if Assigned(FOnRightClick) then FOnRightClick(Self, LL, Params[2].VInteger, Items[Params[2].VInteger]);
  end;
end;

function TGMRectangle.GetAPIUrl: string;
begin
  Result := 'https://developers.google.com/maps/documentation/javascript/reference?hl=en#Rectangle';
end;

function TGMRectangle.GetCollectionClass: TLinkedComponentsClass;
begin
  Result := TRectangles;
end;

function TGMRectangle.GetCollectionItemClass: TLinkedComponentClass;
begin
  Result := TRectangle;
end;

function TGMRectangle.GetItems(I: Integer): TRectangle;
begin
  Result := TRectangle(inherited Items[i]);
end;

procedure TGMRectangle.ShowElements;
const
  StrParams = '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s,%s';
var
  Params: string;
  i: Integer;
begin
  if not ExecuteScript('DeleteObjects', IntToStr(IdxList)) then Exit;

  for i:= 0 to VisualObjects.Count - 1 do
  begin
    Params := Format(StrParams, [
                    IntToStr(IdxList),
                    LowerCase(BoolToStr(Items[i].Clickable, True)),
                    LowerCase(BoolToStr(Items[i].Editable, True)),
                    QuotedStr(Map.TColorToStr(Items[i].FillColor)),
                    StringReplace(FloatToStr(Items[i].FillOpacity), ',', '.', [rfReplaceAll]),
                    QuotedStr(Map.TColorToStr(Items[i].StrokeColor)),
                    StringReplace(FloatToStr(Items[i].StrokeOpacity), ',', '.', [rfReplaceAll]),
                    IntToStr(Items[i].StrokeWeight),
                    LowerCase(BoolToStr(Items[i].Visible, True)),
                    Items[i].Bounds.SW.LatToStr(Map.Precision),
                    Items[i].Bounds.SW.LngToStr(Map.Precision),
                    Items[i].Bounds.NE.LatToStr(Map.Precision),
                    Items[i].Bounds.NE.LngToStr(Map.Precision),
                    IntToStr(Items[i].Index),
                    QuotedStr(Items[i].InfoWindow.GetConvertedString),
                    LowerCase(BoolToStr(Items[i].InfoWindow.DisableAutoPan, True)),
                    IntToStr(Items[i].InfoWindow.MaxWidth),
                    IntToStr(Items[i].InfoWindow.PixelOffset.Height),
                    IntToStr(Items[i].InfoWindow.PixelOffset.Width),
                    LowerCase(BoolToStr(Items[i].InfoWindow.CloseOtherBeforeOpen, True))
                    ]);

    ExecuteScript('MakeRectangle', Params);
  end;
end;

{ TRectangles }

function TRectangles.Add: TRectangle;
begin
  Result := TRectangle(inherited Add);
end;

function TRectangles.GetItems(I: Integer): TRectangle;
begin
  Result := TRectangle(inherited Items[I]);
end;

function TRectangles.GetOwner: TPersistent;
begin
  Result := TGMRectangle(inherited GetOwner);
end;

function TRectangles.Insert(Index: Integer): TRectangle;
begin
  Result := TRectangle(inherited Insert(Index));
end;

procedure TRectangles.SetItems(I: Integer; const Value: TRectangle);
begin
  inherited SetItem(I, Value);
end;

{ TRectangle }

procedure TRectangle.Assign(Source: TPersistent);
begin
  inherited;

  if Source is TRectangle then
  begin
    Bounds.Assign(TRectangle(Source).Bounds);
    Clickable := TRectangle(Source).Clickable;
    Editable  := TRectangle(Source).Editable;
    FillColor := TRectangle(Source).FillColor;
    FillOpacity := TRectangle(Source).FillOpacity;
    StrokeColor := TRectangle(Source).StrokeColor;
    StrokeOpacity := TRectangle(Source).StrokeOpacity;
    StrokeWeight := TRectangle(Source).StrokeWeight;
    Visible := TRectangle(Source).Visible;
  end;
end;

procedure TRectangle.CenterMapTo;
begin
  if Assigned(Collection) and (Collection is TRectangles) and
     Assigned(TRectangles(Collection).FGMLinkedComponent) and
     Assigned(TRectangles(Collection).FGMLinkedComponent.Map) then
    TRectangles(Collection).FGMLinkedComponent.Map.SetCenter(Bounds.GetCenter.Lat, Bounds.GetCenter.Lng);
end;

constructor TRectangle.Create(Collection: TCollection);
begin
  inherited;

  FBounds := TLatLngBounds.Create;
  Bounds.NE.OnChange := OnChangeBounds;
  Bounds.SW.OnChange := OnChangeBounds;
  FClickable := True;
  FEditable := False;
  FFillOpacity := 0.5;
  FFillColor := clRed;
  FStrokeColor := clBlack;
  FStrokeOpacity := 1;
  FStrokeWeight := 2;
  FVisible := True;
  FIsBoundsUpdt := False;
end;

destructor TRectangle.Destroy;
begin
  if Assigned(FBounds) then FreeAndNil(FBounds);

  inherited;
end;

procedure TRectangle.OnChangeBounds(Sender: TObject);
var
  Tmp: string;
begin
  if Assigned(Collection) and (Collection is TRectangles) and
     Assigned(TRectangles(Collection).FGMLinkedComponent) and
     Assigned(TRectangles(Collection).FGMLinkedComponent.Map) then
  begin
    Tmp := Bounds.NE.ToUrlValue(TRectangles(Collection).FGMLinkedComponent.Map.Precision);
    Tmp := Tmp + ',' + Bounds.SW.ToUrlValue(TRectangles(Collection).FGMLinkedComponent.Map.Precision);
    FIsBoundsUpdt := True;
    SetProperty('RectangleSetBounds', 'OnBoundsChanged', Tmp);
  end;
end;

procedure TRectangle.SetClickable(const Value: Boolean);
begin
  if FClickable = Value then Exit;

  FClickable := Value;

  SetProperty('SetClickable', 'OnClickableChange', FClickable);
end;

procedure TRectangle.SetEditable(const Value: Boolean);
begin
  if FEditable = Value then Exit;

  FEditable := Value;

  SetProperty('SetEditable', 'OnEditableChange', FEditable);
end;

procedure TRectangle.SetFillColor(const Value: TColor);
begin
  if FFillColor = Value then Exit;

  FFillColor := Value;

  {$IF CompilerVersion < 20}  // ES: si la versin es inferior a D2009 - EN: if lower than D2009 version
  SetColorProperty('SetFillColor', 'OnFillColorChange', FFillColor);
  {$ELSE}
  SetProperty('SetFillColor', 'OnFillColorChange', FFillColor);
  {$IFEND}
end;

procedure TRectangle.SetFillOpacity(const Value: Real);
begin
  if FFillOpacity = Value then Exit;

  FFillOpacity := Value;

  SetProperty('SetFillOpacity', 'OnFillOpacityChange', FFillOpacity);
end;

procedure TRectangle.SetIdxList(const Value: Cardinal);
begin
  inherited;

end;

procedure TRectangle.SetStrokeColor(const Value: TColor);
begin
  if FStrokeColor = Value then Exit;

  FStrokeColor := Value;

  {$IF CompilerVersion < 20}  // ES: si la versin es inferior a D2009 - EN: if lower than D2009 version
  SetColorProperty('SetStrokeColor', 'OnStrokeColorChange', FStrokeColor);
  {$ELSE}
  SetProperty('SetStrokeColor', 'OnStrokeColorChange', FStrokeColor);
  {$IFEND}
end;

procedure TRectangle.SetStrokeOpacity(const Value: Real);
begin
  if FStrokeOpacity = Value then Exit;

  FStrokeOpacity := Value;

  SetProperty('SetStrokeOpacity', 'OnStrokeOpacityChange', FStrokeOpacity);
end;

procedure TRectangle.SetStrokeWeight(const Value: Integer);
begin
  if FStrokeWeight = Value then Exit;

  FStrokeWeight := Value;

  SetProperty('SetStrokeWeight', 'OnStrokeWeightChange', FStrokeWeight);
end;

procedure TRectangle.SetVisible(const Value: Boolean);
begin
  if FVisible = Value then Exit;

  FVisible := Value;

  SetProperty('SetVisible', 'OnVisibleChange', FVisible);
end;

end.
